
// autogenerated from https://raw.githubusercontent.com/nayarsystems/posix_tz_db/master/zones.csv
// by script <esp8266 arduino core>/tools/TZupdate.sh
// Mon Jul 26 20:04:37 UTC 2021
//
// This database is autogenerated from IANA timezone database
//    https://raw.githubusercontent.com/nayarsystems/posix_tz_db/master/zones.csv
//    (using https://www.iana.org/time-zones)
// and can be updated on demand in this repository
// or by yourself using the above script

#ifndef TZDB_H
#define TZDB_H

#define TZ_Africa_Abidjan	PSTR("GMT0")
#define TZ_Africa_Accra	PSTR("GMT0")
#define TZ_Africa_Addis_Ababa	PSTR("EAT-3")
#define TZ_Africa_Algiers	PSTR("CET-1")
#define TZ_Africa_Asmara	PSTR("EAT-3")
#define TZ_Africa_Bamako	PSTR("GMT0")
#define TZ_Africa_Bangui	PSTR("WAT-1")
#define TZ_Africa_Banjul	PSTR("GMT0")
#define TZ_Africa_Bissau	PSTR("GMT0")
#define TZ_Africa_Blantyre	PSTR("CAT-2")
#define TZ_Africa_Brazzaville	PSTR("WAT-1")
#define TZ_Africa_Bujumbura	PSTR("CAT-2")
#define TZ_Africa_Cairo	PSTR("EET-2")
#define TZ_Africa_Casablanca	PSTR("<+01>-1")
#define TZ_Africa_Ceuta	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Africa_Conakry	PSTR("GMT0")
#define TZ_Africa_Dakar	PSTR("GMT0")
#define TZ_Africa_Dar_es_Salaam	PSTR("EAT-3")
#define TZ_Africa_Djibouti	PSTR("EAT-3")
#define TZ_Africa_Douala	PSTR("WAT-1")
#define TZ_Africa_El_Aaiun	PSTR("<+01>-1")
#define TZ_Africa_Freetown	PSTR("GMT0")
#define TZ_Africa_Gaborone	PSTR("CAT-2")
#define TZ_Africa_Harare	PSTR("CAT-2")
#define TZ_Africa_Johannesburg	PSTR("SAST-2")
#define TZ_Africa_Juba	PSTR("EAT-3")
#define TZ_Africa_Kampala	PSTR("EAT-3")
#define TZ_Africa_Khartoum	PSTR("CAT-2")
#define TZ_Africa_Kigali	PSTR("CAT-2")
#define TZ_Africa_Kinshasa	PSTR("WAT-1")
#define TZ_Africa_Lagos	PSTR("WAT-1")
#define TZ_Africa_Libreville	PSTR("WAT-1")
#define TZ_Africa_Lome	PSTR("GMT0")
#define TZ_Africa_Luanda	PSTR("WAT-1")
#define TZ_Africa_Lubumbashi	PSTR("CAT-2")
#define TZ_Africa_Lusaka	PSTR("CAT-2")
#define TZ_Africa_Malabo	PSTR("WAT-1")
#define TZ_Africa_Maputo	PSTR("CAT-2")
#define TZ_Africa_Maseru	PSTR("SAST-2")
#define TZ_Africa_Mbabane	PSTR("SAST-2")
#define TZ_Africa_Mogadishu	PSTR("EAT-3")
#define TZ_Africa_Monrovia	PSTR("GMT0")
#define TZ_Africa_Nairobi	PSTR("EAT-3")
#define TZ_Africa_Ndjamena	PSTR("WAT-1")
#define TZ_Africa_Niamey	PSTR("WAT-1")
#define TZ_Africa_Nouakchott	PSTR("GMT0")
#define TZ_Africa_Ouagadougou	PSTR("GMT0")
#define TZ_Africa_PortomNovo	PSTR("WAT-1")
#define TZ_Africa_Sao_Tome	PSTR("GMT0")
#define TZ_Africa_Tripoli	PSTR("EET-2")
#define TZ_Africa_Tunis	PSTR("CET-1")
#define TZ_Africa_Windhoek	PSTR("CAT-2")
#define TZ_America_Adak	PSTR("HST10HDT,M3.2.0,M11.1.0")
#define TZ_America_Anchorage	PSTR("AKST9AKDT,M3.2.0,M11.1.0")
#define TZ_America_Anguilla	PSTR("AST4")
#define TZ_America_Antigua	PSTR("AST4")
#define TZ_America_Araguaina	PSTR("<-03>3")
#define TZ_America_Argentina_Buenos_Aires	PSTR("<-03>3")
#define TZ_America_Argentina_Catamarca	PSTR("<-03>3")
#define TZ_America_Argentina_Cordoba	PSTR("<-03>3")
#define TZ_America_Argentina_Jujuy	PSTR("<-03>3")
#define TZ_America_Argentina_La_Rioja	PSTR("<-03>3")
#define TZ_America_Argentina_Mendoza	PSTR("<-03>3")
#define TZ_America_Argentina_Rio_Gallegos	PSTR("<-03>3")
#define TZ_America_Argentina_Salta	PSTR("<-03>3")
#define TZ_America_Argentina_San_Juan	PSTR("<-03>3")
#define TZ_America_Argentina_San_Luis	PSTR("<-03>3")
#define TZ_America_Argentina_Tucuman	PSTR("<-03>3")
#define TZ_America_Argentina_Ushuaia	PSTR("<-03>3")
#define TZ_America_Aruba	PSTR("AST4")
#define TZ_America_Asuncion	PSTR("<-04>4<-03>,M10.1.0/0,M3.4.0/0")
#define TZ_America_Atikokan	PSTR("EST5")
#define TZ_America_Bahia	PSTR("<-03>3")
#define TZ_America_Bahia_Banderas	PSTR("CST6CDT,M4.1.0,M10.5.0")
#define TZ_America_Barbados	PSTR("AST4")
#define TZ_America_Belem	PSTR("<-03>3")
#define TZ_America_Belize	PSTR("CST6")
#define TZ_America_BlancmSablon	PSTR("AST4")
#define TZ_America_Boa_Vista	PSTR("<-04>4")
#define TZ_America_Bogota	PSTR("<-05>5")
#define TZ_America_Boise	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_America_Cambridge_Bay	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_America_Campo_Grande	PSTR("<-04>4")
#define TZ_America_Cancun	PSTR("EST5")
#define TZ_America_Caracas	PSTR("<-04>4")
#define TZ_America_Cayenne	PSTR("<-03>3")
#define TZ_America_Cayman	PSTR("EST5")
#define TZ_America_Chicago	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Chihuahua	PSTR("MST7MDT,M4.1.0,M10.5.0")
#define TZ_America_Costa_Rica	PSTR("CST6")
#define TZ_America_Creston	PSTR("MST7")
#define TZ_America_Cuiaba	PSTR("<-04>4")
#define TZ_America_Curacao	PSTR("AST4")
#define TZ_America_Danmarkshavn	PSTR("GMT0")
#define TZ_America_Dawson	PSTR("MST7")
#define TZ_America_Dawson_Creek	PSTR("MST7")
#define TZ_America_Denver	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_America_Detroit	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Dominica	PSTR("AST4")
#define TZ_America_Edmonton	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_America_Eirunepe	PSTR("<-05>5")
#define TZ_America_El_Salvador	PSTR("CST6")
#define TZ_America_Fortaleza	PSTR("<-03>3")
#define TZ_America_Fort_Nelson	PSTR("MST7")
#define TZ_America_Glace_Bay	PSTR("AST4ADT,M3.2.0,M11.1.0")
#define TZ_America_Godthab	PSTR("<-03>3<-02>,M3.5.0/-2,M10.5.0/-1")
#define TZ_America_Goose_Bay	PSTR("AST4ADT,M3.2.0,M11.1.0")
#define TZ_America_Grand_Turk	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Grenada	PSTR("AST4")
#define TZ_America_Guadeloupe	PSTR("AST4")
#define TZ_America_Guatemala	PSTR("CST6")
#define TZ_America_Guayaquil	PSTR("<-05>5")
#define TZ_America_Guyana	PSTR("<-04>4")
#define TZ_America_Halifax	PSTR("AST4ADT,M3.2.0,M11.1.0")
#define TZ_America_Havana	PSTR("CST5CDT,M3.2.0/0,M11.1.0/1")
#define TZ_America_Hermosillo	PSTR("MST7")
#define TZ_America_Indiana_Indianapolis	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Knox	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Marengo	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Petersburg	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Tell_City	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Vevay	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Vincennes	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Indiana_Winamac	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Inuvik	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_America_Iqaluit	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Jamaica	PSTR("EST5")
#define TZ_America_Juneau	PSTR("AKST9AKDT,M3.2.0,M11.1.0")
#define TZ_America_Kentucky_Louisville	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Kentucky_Monticello	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Kralendijk	PSTR("AST4")
#define TZ_America_La_Paz	PSTR("<-04>4")
#define TZ_America_Lima	PSTR("<-05>5")
#define TZ_America_Los_Angeles	PSTR("PST8PDT,M3.2.0,M11.1.0")
#define TZ_America_Lower_Princes	PSTR("AST4")
#define TZ_America_Maceio	PSTR("<-03>3")
#define TZ_America_Managua	PSTR("CST6")
#define TZ_America_Manaus	PSTR("<-04>4")
#define TZ_America_Marigot	PSTR("AST4")
#define TZ_America_Martinique	PSTR("AST4")
#define TZ_America_Matamoros	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Mazatlan	PSTR("MST7MDT,M4.1.0,M10.5.0")
#define TZ_America_Menominee	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Merida	PSTR("CST6CDT,M4.1.0,M10.5.0")
#define TZ_America_Metlakatla	PSTR("AKST9AKDT,M3.2.0,M11.1.0")
#define TZ_America_Mexico_City	PSTR("CST6CDT,M4.1.0,M10.5.0")
#define TZ_America_Miquelon	PSTR("<-03>3<-02>,M3.2.0,M11.1.0")
#define TZ_America_Moncton	PSTR("AST4ADT,M3.2.0,M11.1.0")
#define TZ_America_Monterrey	PSTR("CST6CDT,M4.1.0,M10.5.0")
#define TZ_America_Montevideo	PSTR("<-03>3")
#define TZ_America_Montreal	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Montserrat	PSTR("AST4")
#define TZ_America_Nassau	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_New_York	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Nipigon	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Nome	PSTR("AKST9AKDT,M3.2.0,M11.1.0")
#define TZ_America_Noronha	PSTR("<-02>2")
#define TZ_America_North_Dakota_Beulah	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_North_Dakota_Center	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_North_Dakota_New_Salem	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Nuuk	PSTR("<-03>3<-02>,M3.5.0/-2,M10.5.0/-1")
#define TZ_America_Ojinaga	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_America_Panama	PSTR("EST5")
#define TZ_America_Pangnirtung	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Paramaribo	PSTR("<-03>3")
#define TZ_America_Phoenix	PSTR("MST7")
#define TZ_America_PortmaumPrince	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Port_of_Spain	PSTR("AST4")
#define TZ_America_Porto_Velho	PSTR("<-04>4")
#define TZ_America_Puerto_Rico	PSTR("AST4")
#define TZ_America_Punta_Arenas	PSTR("<-03>3")
#define TZ_America_Rainy_River	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Rankin_Inlet	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Recife	PSTR("<-03>3")
#define TZ_America_Regina	PSTR("CST6")
#define TZ_America_Resolute	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Rio_Branco	PSTR("<-05>5")
#define TZ_America_Santarem	PSTR("<-03>3")
#define TZ_America_Santiago	PSTR("<-04>4<-03>,M9.1.6/24,M4.1.6/24")
#define TZ_America_Santo_Domingo	PSTR("AST4")
#define TZ_America_Sao_Paulo	PSTR("<-03>3")
#define TZ_America_Scoresbysund	PSTR("<-01>1<+00>,M3.5.0/0,M10.5.0/1")
#define TZ_America_Sitka	PSTR("AKST9AKDT,M3.2.0,M11.1.0")
#define TZ_America_St_Barthelemy	PSTR("AST4")
#define TZ_America_St_Johns	PSTR("NST3:30NDT,M3.2.0,M11.1.0")
#define TZ_America_St_Kitts	PSTR("AST4")
#define TZ_America_St_Lucia	PSTR("AST4")
#define TZ_America_St_Thomas	PSTR("AST4")
#define TZ_America_St_Vincent	PSTR("AST4")
#define TZ_America_Swift_Current	PSTR("CST6")
#define TZ_America_Tegucigalpa	PSTR("CST6")
#define TZ_America_Thule	PSTR("AST4ADT,M3.2.0,M11.1.0")
#define TZ_America_Thunder_Bay	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Tijuana	PSTR("PST8PDT,M3.2.0,M11.1.0")
#define TZ_America_Toronto	PSTR("EST5EDT,M3.2.0,M11.1.0")
#define TZ_America_Tortola	PSTR("AST4")
#define TZ_America_Vancouver	PSTR("PST8PDT,M3.2.0,M11.1.0")
#define TZ_America_Whitehorse	PSTR("MST7")
#define TZ_America_Winnipeg	PSTR("CST6CDT,M3.2.0,M11.1.0")
#define TZ_America_Yakutat	PSTR("AKST9AKDT,M3.2.0,M11.1.0")
#define TZ_America_Yellowknife	PSTR("MST7MDT,M3.2.0,M11.1.0")
#define TZ_Antarctica_Casey	PSTR("<+11>-11")
#define TZ_Antarctica_Davis	PSTR("<+07>-7")
#define TZ_Antarctica_DumontDUrville	PSTR("<+10>-10")
#define TZ_Antarctica_Macquarie	PSTR("AEST-10AEDT,M10.1.0,M4.1.0/3")
#define TZ_Antarctica_Mawson	PSTR("<+05>-5")
#define TZ_Antarctica_McMurdo	PSTR("NZST-12NZDT,M9.5.0,M4.1.0/3")
#define TZ_Antarctica_Palmer	PSTR("<-03>3")
#define TZ_Antarctica_Rothera	PSTR("<-03>3")
#define TZ_Antarctica_Syowa	PSTR("<+03>-3")
#define TZ_Antarctica_Troll	PSTR("<+00>0<+02>-2,M3.5.0/1,M10.5.0/3")
#define TZ_Antarctica_Vostok	PSTR("<+06>-6")
#define TZ_Arctic_Longyearbyen	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Asia_Aden	PSTR("<+03>-3")
#define TZ_Asia_Almaty	PSTR("<+06>-6")
#define TZ_Asia_Amman	PSTR("EET-2EEST,M3.5.4/24,M10.5.5/1")
#define TZ_Asia_Anadyr	PSTR("<+12>-12")
#define TZ_Asia_Aqtau	PSTR("<+05>-5")
#define TZ_Asia_Aqtobe	PSTR("<+05>-5")
#define TZ_Asia_Ashgabat	PSTR("<+05>-5")
#define TZ_Asia_Atyrau	PSTR("<+05>-5")
#define TZ_Asia_Baghdad	PSTR("<+03>-3")
#define TZ_Asia_Bahrain	PSTR("<+03>-3")
#define TZ_Asia_Baku	PSTR("<+04>-4")
#define TZ_Asia_Bangkok	PSTR("<+07>-7")
#define TZ_Asia_Barnaul	PSTR("<+07>-7")
#define TZ_Asia_Beirut	PSTR("EET-2EEST,M3.5.0/0,M10.5.0/0")
#define TZ_Asia_Bishkek	PSTR("<+06>-6")
#define TZ_Asia_Brunei	PSTR("<+08>-8")
#define TZ_Asia_Chita	PSTR("<+09>-9")
#define TZ_Asia_Choibalsan	PSTR("<+08>-8")
#define TZ_Asia_Colombo	PSTR("<+0530>-5:30")
#define TZ_Asia_Damascus	PSTR("EET-2EEST,M3.5.5/0,M10.5.5/0")
#define TZ_Asia_Dhaka	PSTR("<+06>-6")
#define TZ_Asia_Dili	PSTR("<+09>-9")
#define TZ_Asia_Dubai	PSTR("<+04>-4")
#define TZ_Asia_Dushanbe	PSTR("<+05>-5")
#define TZ_Asia_Famagusta	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Asia_Gaza	PSTR("EET-2EEST,M3.4.4/48,M10.4.4/49")
#define TZ_Asia_Hebron	PSTR("EET-2EEST,M3.4.4/48,M10.4.4/49")
#define TZ_Asia_Ho_Chi_Minh	PSTR("<+07>-7")
#define TZ_Asia_Hong_Kong	PSTR("HKT-8")
#define TZ_Asia_Hovd	PSTR("<+07>-7")
#define TZ_Asia_Irkutsk	PSTR("<+08>-8")
#define TZ_Asia_Jakarta	PSTR("WIB-7")
#define TZ_Asia_Jayapura	PSTR("WIT-9")
#define TZ_Asia_Jerusalem	PSTR("IST-2IDT,M3.4.4/26,M10.5.0")
#define TZ_Asia_Kabul	PSTR("<+0430>-4:30")
#define TZ_Asia_Kamchatka	PSTR("<+12>-12")
#define TZ_Asia_Karachi	PSTR("PKT-5")
#define TZ_Asia_Kathmandu	PSTR("<+0545>-5:45")
#define TZ_Asia_Khandyga	PSTR("<+09>-9")
#define TZ_Asia_Kolkata	PSTR("IST-5:30")
#define TZ_Asia_Krasnoyarsk	PSTR("<+07>-7")
#define TZ_Asia_Kuala_Lumpur	PSTR("<+08>-8")
#define TZ_Asia_Kuching	PSTR("<+08>-8")
#define TZ_Asia_Kuwait	PSTR("<+03>-3")
#define TZ_Asia_Macau	PSTR("CST-8")
#define TZ_Asia_Magadan	PSTR("<+11>-11")
#define TZ_Asia_Makassar	PSTR("WITA-8")
#define TZ_Asia_Manila	PSTR("PST-8")
#define TZ_Asia_Muscat	PSTR("<+04>-4")
#define TZ_Asia_Nicosia	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Asia_Novokuznetsk	PSTR("<+07>-7")
#define TZ_Asia_Novosibirsk	PSTR("<+07>-7")
#define TZ_Asia_Omsk	PSTR("<+06>-6")
#define TZ_Asia_Oral	PSTR("<+05>-5")
#define TZ_Asia_Phnom_Penh	PSTR("<+07>-7")
#define TZ_Asia_Pontianak	PSTR("WIB-7")
#define TZ_Asia_Pyongyang	PSTR("KST-9")
#define TZ_Asia_Qatar	PSTR("<+03>-3")
#define TZ_Asia_Qyzylorda	PSTR("<+05>-5")
#define TZ_Asia_Riyadh	PSTR("<+03>-3")
#define TZ_Asia_Sakhalin	PSTR("<+11>-11")
#define TZ_Asia_Samarkand	PSTR("<+05>-5")
#define TZ_Asia_Seoul	PSTR("KST-9")
#define TZ_Asia_Shanghai	PSTR("CST-8")
#define TZ_Asia_Singapore	PSTR("<+08>-8")
#define TZ_Asia_Srednekolymsk	PSTR("<+11>-11")
#define TZ_Asia_Taipei	PSTR("CST-8")
#define TZ_Asia_Tashkent	PSTR("<+05>-5")
#define TZ_Asia_Tbilisi	PSTR("<+04>-4")
#define TZ_Asia_Tehran	PSTR("<+0330>-3:30<+0430>,J79/24,J263/24")
#define TZ_Asia_Thimphu	PSTR("<+06>-6")
#define TZ_Asia_Tokyo	PSTR("JST-9")
#define TZ_Asia_Tomsk	PSTR("<+07>-7")
#define TZ_Asia_Ulaanbaatar	PSTR("<+08>-8")
#define TZ_Asia_Urumqi	PSTR("<+06>-6")
#define TZ_Asia_UstmNera	PSTR("<+10>-10")
#define TZ_Asia_Vientiane	PSTR("<+07>-7")
#define TZ_Asia_Vladivostok	PSTR("<+10>-10")
#define TZ_Asia_Yakutsk	PSTR("<+09>-9")
#define TZ_Asia_Yangon	PSTR("<+0630>-6:30")
#define TZ_Asia_Yekaterinburg	PSTR("<+05>-5")
#define TZ_Asia_Yerevan	PSTR("<+04>-4")
#define TZ_Atlantic_Azores	PSTR("<-01>1<+00>,M3.5.0/0,M10.5.0/1")
#define TZ_Atlantic_Bermuda	PSTR("AST4ADT,M3.2.0,M11.1.0")
#define TZ_Atlantic_Canary	PSTR("WET0WEST,M3.5.0/1,M10.5.0")
#define TZ_Atlantic_Cape_Verde	PSTR("<-01>1")
#define TZ_Atlantic_Faroe	PSTR("WET0WEST,M3.5.0/1,M10.5.0")
#define TZ_Atlantic_Madeira	PSTR("WET0WEST,M3.5.0/1,M10.5.0")
#define TZ_Atlantic_Reykjavik	PSTR("GMT0")
#define TZ_Atlantic_South_Georgia	PSTR("<-02>2")
#define TZ_Atlantic_Stanley	PSTR("<-03>3")
#define TZ_Atlantic_St_Helena	PSTR("GMT0")
#define TZ_Australia_Adelaide	PSTR("ACST-9:30ACDT,M10.1.0,M4.1.0/3")
#define TZ_Australia_Brisbane	PSTR("AEST-10")
#define TZ_Australia_Broken_Hill	PSTR("ACST-9:30ACDT,M10.1.0,M4.1.0/3")
#define TZ_Australia_Currie	PSTR("AEST-10AEDT,M10.1.0,M4.1.0/3")
#define TZ_Australia_Darwin	PSTR("ACST-9:30")
#define TZ_Australia_Eucla	PSTR("<+0845>-8:45")
#define TZ_Australia_Hobart	PSTR("AEST-10AEDT,M10.1.0,M4.1.0/3")
#define TZ_Australia_Lindeman	PSTR("AEST-10")
#define TZ_Australia_Lord_Howe	PSTR("<+1030>-10:30<+11>-11,M10.1.0,M4.1.0")
#define TZ_Australia_Melbourne	PSTR("AEST-10AEDT,M10.1.0,M4.1.0/3")
#define TZ_Australia_Perth	PSTR("AWST-8")
#define TZ_Australia_Sydney	PSTR("AEST-10AEDT,M10.1.0,M4.1.0/3")
#define TZ_Europe_Amsterdam	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Andorra	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Astrakhan	PSTR("<+04>-4")
#define TZ_Europe_Athens	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Belgrade	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Berlin	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Bratislava	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Brussels	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Bucharest	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Budapest	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Busingen	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Chisinau	PSTR("EET-2EEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Copenhagen	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Dublin	PSTR("IST-1GMT0,M10.5.0,M3.5.0/1")
#define TZ_Europe_Gibraltar	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Guernsey	PSTR("GMT0BST,M3.5.0/1,M10.5.0")
#define TZ_Europe_Helsinki	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Isle_of_Man	PSTR("GMT0BST,M3.5.0/1,M10.5.0")
#define TZ_Europe_Istanbul	PSTR("<+03>-3")
#define TZ_Europe_Jersey	PSTR("GMT0BST,M3.5.0/1,M10.5.0")
#define TZ_Europe_Kaliningrad	PSTR("EET-2")
#define TZ_Europe_Kiev	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Kirov	PSTR("<+03>-3")
#define TZ_Europe_Lisbon	PSTR("WET0WEST,M3.5.0/1,M10.5.0")
#define TZ_Europe_Ljubljana	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_London	PSTR("GMT0BST,M3.5.0/1,M10.5.0")
#define TZ_Europe_Luxembourg	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Madrid	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Malta	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Mariehamn	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Minsk	PSTR("<+03>-3")
#define TZ_Europe_Monaco	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Moscow	PSTR("MSK-3")
#define TZ_Europe_Oslo	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Paris	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Podgorica	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Prague	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Riga	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Rome	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Samara	PSTR("<+04>-4")
#define TZ_Europe_San_Marino	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Sarajevo	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Saratov	PSTR("<+04>-4")
#define TZ_Europe_Simferopol	PSTR("MSK-3")
#define TZ_Europe_Skopje	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Sofia	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Stockholm	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Tallinn	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Tirane	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Ulyanovsk	PSTR("<+04>-4")
#define TZ_Europe_Uzhgorod	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Vaduz	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Vatican	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Vienna	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Vilnius	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Volgograd	PSTR("<+03>-3")
#define TZ_Europe_Warsaw	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Zagreb	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Europe_Zaporozhye	PSTR("EET-2EEST,M3.5.0/3,M10.5.0/4")
#define TZ_Europe_Zurich	PSTR("CET-1CEST,M3.5.0,M10.5.0/3")
#define TZ_Indian_Antananarivo	PSTR("EAT-3")
#define TZ_Indian_Chagos	PSTR("<+06>-6")
#define TZ_Indian_Christmas	PSTR("<+07>-7")
#define TZ_Indian_Cocos	PSTR("<+0630>-6:30")
#define TZ_Indian_Comoro	PSTR("EAT-3")
#define TZ_Indian_Kerguelen	PSTR("<+05>-5")
#define TZ_Indian_Mahe	PSTR("<+04>-4")
#define TZ_Indian_Maldives	PSTR("<+05>-5")
#define TZ_Indian_Mauritius	PSTR("<+04>-4")
#define TZ_Indian_Mayotte	PSTR("EAT-3")
#define TZ_Indian_Reunion	PSTR("<+04>-4")
#define TZ_Pacific_Apia	PSTR("<+13>-13<+14>,M9.5.0/3,M4.1.0/4")
#define TZ_Pacific_Auckland	PSTR("NZST-12NZDT,M9.5.0,M4.1.0/3")
#define TZ_Pacific_Bougainville	PSTR("<+11>-11")
#define TZ_Pacific_Chatham	PSTR("<+1245>-12:45<+1345>,M9.5.0/2:45,M4.1.0/3:45")
#define TZ_Pacific_Chuuk	PSTR("<+10>-10")
#define TZ_Pacific_Easter	PSTR("<-06>6<-05>,M9.1.6/22,M4.1.6/22")
#define TZ_Pacific_Efate	PSTR("<+11>-11")
#define TZ_Pacific_Enderbury	PSTR("<+13>-13")
#define TZ_Pacific_Fakaofo	PSTR("<+13>-13")
#define TZ_Pacific_Fiji	PSTR("<+12>-12<+13>,M11.2.0,M1.2.3/99")
#define TZ_Pacific_Funafuti	PSTR("<+12>-12")
#define TZ_Pacific_Galapagos	PSTR("<-06>6")
#define TZ_Pacific_Gambier	PSTR("<-09>9")
#define TZ_Pacific_Guadalcanal	PSTR("<+11>-11")
#define TZ_Pacific_Guam	PSTR("ChST-10")
#define TZ_Pacific_Honolulu	PSTR("HST10")
#define TZ_Pacific_Kiritimati	PSTR("<+14>-14")
#define TZ_Pacific_Kosrae	PSTR("<+11>-11")
#define TZ_Pacific_Kwajalein	PSTR("<+12>-12")
#define TZ_Pacific_Majuro	PSTR("<+12>-12")
#define TZ_Pacific_Marquesas	PSTR("<-0930>9:30")
#define TZ_Pacific_Midway	PSTR("SST11")
#define TZ_Pacific_Nauru	PSTR("<+12>-12")
#define TZ_Pacific_Niue	PSTR("<-11>11")
#define TZ_Pacific_Norfolk	PSTR("<+11>-11<+12>,M10.1.0,M4.1.0/3")
#define TZ_Pacific_Noumea	PSTR("<+11>-11")
#define TZ_Pacific_Pago_Pago	PSTR("SST11")
#define TZ_Pacific_Palau	PSTR("<+09>-9")
#define TZ_Pacific_Pitcairn	PSTR("<-08>8")
#define TZ_Pacific_Pohnpei	PSTR("<+11>-11")
#define TZ_Pacific_Port_Moresby	PSTR("<+10>-10")
#define TZ_Pacific_Rarotonga	PSTR("<-10>10")
#define TZ_Pacific_Saipan	PSTR("ChST-10")
#define TZ_Pacific_Tahiti	PSTR("<-10>10")
#define TZ_Pacific_Tarawa	PSTR("<+12>-12")
#define TZ_Pacific_Tongatapu	PSTR("<+13>-13")
#define TZ_Pacific_Wake	PSTR("<+12>-12")
#define TZ_Pacific_Wallis	PSTR("<+12>-12")
#define TZ_Etc_GMT	PSTR("GMT0")
#define TZ_Etc_GMTm0	PSTR("GMT0")
#define TZ_Etc_GMTm1	PSTR("<+01>-1")
#define TZ_Etc_GMTm2	PSTR("<+02>-2")
#define TZ_Etc_GMTm3	PSTR("<+03>-3")
#define TZ_Etc_GMTm4	PSTR("<+04>-4")
#define TZ_Etc_GMTm5	PSTR("<+05>-5")
#define TZ_Etc_GMTm6	PSTR("<+06>-6")
#define TZ_Etc_GMTm7	PSTR("<+07>-7")
#define TZ_Etc_GMTm8	PSTR("<+08>-8")
#define TZ_Etc_GMTm9	PSTR("<+09>-9")
#define TZ_Etc_GMTm10	PSTR("<+10>-10")
#define TZ_Etc_GMTm11	PSTR("<+11>-11")
#define TZ_Etc_GMTm12	PSTR("<+12>-12")
#define TZ_Etc_GMTm13	PSTR("<+13>-13")
#define TZ_Etc_GMTm14	PSTR("<+14>-14")
#define TZ_Etc_GMT0	PSTR("GMT0")
#define TZ_Etc_GMTp0	PSTR("GMT0")
#define TZ_Etc_GMTp1	PSTR("<-01>1")
#define TZ_Etc_GMTp2	PSTR("<-02>2")
#define TZ_Etc_GMTp3	PSTR("<-03>3")
#define TZ_Etc_GMTp4	PSTR("<-04>4")
#define TZ_Etc_GMTp5	PSTR("<-05>5")
#define TZ_Etc_GMTp6	PSTR("<-06>6")
#define TZ_Etc_GMTp7	PSTR("<-07>7")
#define TZ_Etc_GMTp8	PSTR("<-08>8")
#define TZ_Etc_GMTp9	PSTR("<-09>9")
#define TZ_Etc_GMTp10	PSTR("<-10>10")
#define TZ_Etc_GMTp11	PSTR("<-11>11")
#define TZ_Etc_GMTp12	PSTR("<-12>12")
#define TZ_Etc_UCT	PSTR("UTC0")
#define TZ_Etc_UTC	PSTR("UTC0")
#define TZ_Etc_Greenwich	PSTR("GMT0")
#define TZ_Etc_Universal	PSTR("UTC0")
#define TZ_Etc_Zulu	PSTR("UTC0")

#endif // TZDB_H
