#ifndef LeonardoGpioPinMap_h
#define LeonardoGpioPinMap_h


namespace sdfat {


static const GpioPinMap_t GpioPinMap[] = {
  GPIO_PIN(D, 2),  // D0
  GPIO_PIN(D, 3),  // D1
  GPIO_PIN(D, 1),  // D2
  GPIO_PIN(D, 0),  // D3
  GPIO_PIN(D, 4),  // D4
  GPIO_PIN(C, 6),  // D5
  GPIO_PIN(D, 7),  // D6
  GPIO_PIN(E, 6),  // D7
  GPIO_PIN(B, 4),  // D8
  GPIO_PIN(B, 5),  // D9
  GPIO_PIN(B, 6),  // D10
  GPIO_PIN(B, 7),  // D11
  GPIO_PIN(D, 6),  // D12
  GPIO_PIN(C, 7),  // D13
  GPIO_PIN(B, 3),  // D14
  GPIO_PIN(B, 1),  // D15
  GPIO_PIN(B, 2),  // D16
  GPIO_PIN(B, 0),  // D17
  GPIO_PIN(F, 7),  // D18
  GPIO_PIN(F, 6),  // D19
  GPIO_PIN(F, 5),  // D20
  GPIO_PIN(F, 4),  // D21
  GPIO_PIN(F, 1),  // D22
  GPIO_PIN(F, 0),  // D23
  GPIO_PIN(D, 4),  // D24
  GPIO_PIN(D, 7),  // D25
  GPIO_PIN(B, 4),  // D26
  GPIO_PIN(B, 5),  // D27
  GPIO_PIN(B, 6),  // D28
  GPIO_PIN(D, 6)   // D29
};


}; // namespace sdfat


#endif  // LeonardoGpioPinMap_h
